# Copyright 2018 Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from py4j.java_gateway import JavaObject
from py4j.java_collections import JavaList

from pyspark3d import load_from_jvm
from pyspark3d import get_spark_context

from pyspark import RDD
from pyspark.mllib.common import _java2py

def scala2java(scala_list: JavaObject) -> JavaList:
    """
    Convert a Scala list coming from the JVM into a JavaList:
    scala.collection.immutable.List -> java.util.ArrayList

    Parameters
    ----------
    scala_list : JavaObject
        Scala list (scala.collection.immutable.List)

    Returns
    ----------
    java_list : JavaList
        Java list (java.util.ArrayList)

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to Java List
    >>> coord_java = scala2java(coord_scala)
    >>> assert(type(coord_java) == JavaList)
    """
    # Not very powerful check... As many things other than Scala list
    # are JavaObject...
    msg = """
    converters.scala2java expects a JavaObject not {}.
    """.format(type(scala_list))
    assert(type(scala_list) == JavaObject), msg

    convpath = "scala.collection.JavaConverters.seqAsJavaListConverter"
    converter = load_from_jvm(convpath)
    return converter(scala_list).asJava()

def java2python(java_list: JavaList) -> list:
    """
    Convert a Java list into a python list:
    java.util.ArrayList -> list

    Parameters
    ----------
    java_list : JavaList
        Java list (java.util.ArrayList)

    Returns
    ----------
    python_list : list
        Python list

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to Java List
    >>> coord_java = scala2java(coord_scala)
    >>> assert(type(coord_java) == JavaList)

    Convert then the Java List into a python list
    >>> coord_python = java2python(coord_java)
    >>> assert(type(coord_python) == list)

    Other example. Instantiate directly a JavaList in python
    >>> java_list = sc._jvm.java.util.ArrayList()
    >>> java_list.add(0.0) # return a boolean
    True

    >>> python_list = java2python(java_list)
    >>> assert(python_list[0] == 0.0)
    """
    msg = """
    converters.java2python expects a JavaList not {}.
    """.format(type(java_list))
    assert(type(java_list) == JavaList), msg

    return list(java_list)

def scala2python(scala_list: JavaObject) -> list:
    """
    Convert a Scala list coming from the JVM into a python list:
    scala.collection.immutable.List -> list

    Parameters
    ----------
    scala_list : JavaObject
        Scala list (scala.collection.immutable.List)

    Returns
    ----------
    python_list : list
        Python list

    Examples
    ----------
    Instantiate a Point3D
    >>> from pyspark3d.geometryObjects import Point3D
    >>> p = Point3D(0.0, 0.0, 0.0, True)

    Grab the coordinates
    >>> coord_scala = p.getCoordinate()
    >>> assert(type(coord_scala) == JavaObject)

    `coord_scala` is a Scala List. Convert it to python List
    >>> coord_python = scala2python(coord_scala)
    >>> assert(type(coord_python) == list)
    """
    # Not very powerful check... As many things other than Scala list
    # are JavaObject...
    msg = """
    converters.scala2python expects a JavaObject not {}.
    """.format(type(scala_list))
    assert(type(scala_list) == JavaObject), msg

    return java2python(scala2java(scala_list))

def toCoordRDD(
        srdd: JavaObject, gridtype: str="", numPartitions: int=None) -> RDD:
    """Convert a RDD of Shape3D objects from spark3D into a PythonRDD whose
    elements are the coordinates of the Shape3D objects.

    The element of a RDD coming from the Scala/Java world won't be usable in
    general in Python as they aren't defined (unless you wrote explicitly the
    converter). For example, a RDD[Point3D] is understood in Python, but
    you won't be able to manipulate its elements (Point3D are Java objects).
    The idea is then to manipulate the full RDD in Scala, but interface just
    the coordinates in the end, e.g. for visualisation.

    By default, `toCoord` will act on the raw RDD. You can also repartition the
    RDD before the conversion.

    Parameters
    ----------
    srdd : JavaObject
        Point3DRDD or SphereRDD instance (spatial3DRDD).
    gridtype : str, optional
        Type of the repartitioning to apply: LINEARONIONGRID, OCTREE. Default
        is no repartitioning (gridtype="").
    numPartitions : int, optional
        Number of partitions after repartitioning.

    Returns
    -------
    RDD
        PythonRDD whose elements are object centers. If gridtype is specified,
        the RDD has been repartitioned.

    Examples
    -------
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import Point3DRDD

    Load data
    >>> fn = os.path.join(path_to_conf, "../src/test/resources/astro_obs.fits")
    >>> p3d = Point3DRDD(spark, fn, "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    No repartitioning & no change of partition number
    >>> pyrdd = toCoordRDD(p3d)
    >>> print(round(pyrdd.first()[0], 2))
    0.55

    No repartitioning but increase the number of partitions
    >>> pyrdd = toCoordRDD(p3d, numPartitions=100)
    >>> print(round(pyrdd.first()[0], 2))
    0.06
    >>> print(pyrdd.getNumPartitions())
    100

    OCTREE repartitioning, and increase the number of partitions
    >>> pyrdd = toCoordRDD(p3d, "OCTREE", 100)
    >>> print(round(pyrdd.first()[0], 2))
    0.92

    For Octree, the number of partition is always a power of 8.
    In this case, 8**2 is the closest.
    >>> print(pyrdd.getNumPartitions())
    64

    """
    pysc = get_spark_context()

    # Get the desired final number of partitions
    if numPartitions is None:
        npart = srdd.rawRDD().getNumPartitions()
    else:
        npart = numPartitions

    # Repartition if needed
    if gridtype != "":
        rdd = srdd.spatialPartitioningPython(gridtype, npart)
    else:
        if numPartitions is None:
            rdd = srdd.rawRDD()
        else:
            rdd = srdd.rawRDD()
            return _java2py(
                pysc, srdd.toCenterCoordinateRDDPython(rdd)).repartition(npart)
    return _java2py(pysc, srdd.toCenterCoordinateRDDPython(rdd))


if __name__ == "__main__":
    """
    Run the doctest using

    python converters.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import os
    import sys
    import doctest
    import numpy as np

    from pyspark import SparkContext
    from pyspark3d import pyspark3d_conf
    from pyspark3d import load_user_conf
    from pyspark3d import get_spark_session

    # Activate the SparkContext for the test suite
    dic = load_user_conf()
    conf = pyspark3d_conf("local[*]", "test", dic)
    sc = SparkContext.getOrCreate(conf=conf)

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    spark = get_spark_session(dicconf=dic)

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    # Run the test suite
    failure_count, test_count = doctest.testmod(extraglobs={"spark": spark})
    sys.exit(failure_count)



# Copyright 2018 Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pyspark.sql import SparkSession

from py4j.java_gateway import JavaObject

from pyspark3d import load_from_jvm
from pyspark3d.pyspark3d_conf import path_to_conf

import os
from typing import Dict

def Point3DRDD(
        spark: SparkSession, filename: str, colnames: str,
        isSpherical: bool, format: str, options: Dict={"": ""}) -> JavaObject:
    """
    Binding around Point3DRDD.scala. For full description,
    see `$spark3d/src/main/scala/com/spark3d/spatial3DRDD/Point3DRDD.scala`

    Construct a Point3DRDD from a RDD[Point3D].
    The RDD[Point3D] is constructed from any available data source registered
    in Spark. For more information about available official connectors:
    `https://spark-packages.org/?q=tags%3A%22Data%20Sources%22`.
    That currently includes: CSV, JSON, TXT, FITS, ROOT, HDF5, Avro, Parquet...

    Note
    -----------
    The Scala version makes use of several constructors (i.e. with different
    kinds of argument). Here we only provide one way to instantiate a
    Point3DRDD Scala class, through the full list of arguments.

    Note that pyspark works with Python wrappers around the *Java* version
    of Spark objects, not around the *Scala* version of Spark objects.
    Therefore on the Scala side, we trigger the method
    `Point3DRDDFromV2PythonHelper` which is a modified version of
    `Point3DRDDFromV2`. The main change is that `options` on the Scala side
    is a java.util.HashMap in order to smoothly connect to `dictionary` in
    the Python side.

    Also for convenience (for the developper!), the StorageLevel is no more an
    argument in the constructor but is set to `StorageLevel.MEMORY_ONLY`.
    This is because I couldn't find how to pass that information from
    Python to Java... TBD!

    Parameters
    ----------
    spark : SparkSession
        The current Spark session.
    filename : str
        File name where the data is stored.
    colnames : str
        Comma-separated names of (x,y,z) columns. Example: "Z_COSMO,RA,Dec".
    isSpherical : bool
        If true, it assumes that the coordinates of the Point3D are
        (r, theta, phi). Otherwise, it assumes cartesian coordinates (x, y, z).
    format : str
        The name of the data source as registered in Spark. For example:
            - text
            - csv
            - json
            - com.astrolabsoftware.sparkfits or fits
            - org.dianahep.sparkroot
            - gov.llnl.spark.hdf or hdf5
    options : dic of (str: str), optional
        Options to pass to the DataFrameReader. Default is no options.

    Returns
    ----------
    p3drdd : Point3DRDD instance
        Instance of the Scala Point3DRDD class.

    Examples
    ----------
    Load data
    >>> fn = os.path.join(path_to_conf, "../src/test/resources/astro_obs.fits")
    >>> rdd = Point3DRDD(spark, fn, "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    Check we have a spatial3DRDD
    >>> assert("com.astrolabsoftware.spark3d.spatial3DRDD" in rdd.toString())

    Count the number of elements
    >>> print(rdd.rawRDD().count())
    20000

    Repartition the data of the RDD (ONION)
    >>> gridtype = "LINEARONIONGRID"
    >>> rdd_part = rdd.spatialPartitioningPython(gridtype,
    ...     rdd.rawRDD().getNumPartitions())

    Repartition the data of the RDD (OCTREE)
    >>> gridtype = "OCTREE"
    >>> rdd_part = rdd.spatialPartitioningPython(gridtype,
    ...     rdd.rawRDD().getNumPartitions())

    Get a RDD with the coordinates of the Point3D
    centers (e.g. useful for plot)
    >>> rdd_centers = rdd.toCenterCoordinateRDDPython(rdd.rawRDD())
    >>> print(round(list(rdd_centers.first())[0], 2))
    0.55

    To see all the available methods:
    >>> print(sorted(rdd.__dir__())) # doctest: +NORMALIZE_WHITESPACE
    ['$lessinit$greater$default$6', '$lessinit$greater$default$7', 'apply',
    'boundary', 'boundary_$eq',
    'com$astrolabsoftware$spark3d$spatial3DRDD$Shape3DRDD$$mapElements$1',
    'equals', 'getClass', 'getDataEnvelope', 'hashCode', 'isSpherical',
    'notify', 'notifyAll', 'partition', 'rawRDD', 'spatialPartitioning',
    'spatialPartitioning$default$2', 'spatialPartitioningPython',
    'spatialPartitioningPython$default$2', 'toCenterCoordinateRDD',
    'toCenterCoordinateRDDPython', 'toString', 'wait']

    """
    scalapath = "com.astrolabsoftware.spark3d.spatial3DRDD.Point3DRDD"
    scalaclass = load_from_jvm(scalapath)

    p3drdd = scalaclass(
        spark._jwrapped.sparkSession(), filename, colnames,
        isSpherical, format, options)

    return p3drdd

def SphereRDD(
        spark: SparkSession, filename: str, colnames: str,
        isSpherical: bool, format: str, options: Dict={"": ""}) -> JavaObject:
    """
    Binding around SphereRDD.scala. For full description,
    see `$spark3d/src/main/scala/com/spark3d/spatial3DRDD/SphereRDD.scala`

    Construct a SphereRDD from a RDD[ShellEnvelope].
    The RDD[Point3D] is constructed from any available data source registered
    in Spark. For more information about available official connectors:
    `https://spark-packages.org/?q=tags%3A%22Data%20Sources%22`.
    That currently includes: CSV, JSON, TXT, FITS, ROOT, HDF5, Avro, Parquet...

    Note
    -----------
    The Scala version makes use of several constructors (i.e. with different
    kinds of argument). Here we only provide one way to instantiate a
    SphereRDD Scala class, through the full list of arguments.

    Note that pyspark works with Python wrappers around the *Java* version
    of Spark objects, not around the *Scala* version of Spark objects.
    Therefore on the Scala side, we trigger the method
    `SphereRDDFromV2PythonHelper` which is a modified version of
    `SphereRDDFromV2`. The main change is that `options` on the Scala side
    is a java.util.HashMap in order to smoothly connect to `dictionary` in
    the Python side.

    Also for convenience (for the developper!), the StorageLevel is no more an
    argument in the constructor but is set to `StorageLevel.MEMORY_ONLY`.
    This is because I couldn't find how to pass that information from
    Python to Java... TBD!

    Parameters
    ----------
    spark : SparkSession
        The current Spark session.
    filename : str
        File name where the data is stored.
    colnames : str
        Comma-separated names of (x,y,z,radius) columns.
        Example: "Z_COSMO,RA,Dec,Radius".
    isSpherical : bool
        If true, it assumes that the coordinates of the Point3D are
        (r, theta, phi). Otherwise, it assumes cartesian coordinates (x, y, z).
    format : str
        The name of the data source as registered in Spark. For example:
            - text
            - csv
            - json
            - com.astrolabsoftware.sparkfits or fits
            - org.dianahep.sparkroot
            - gov.llnl.spark.hdf or hdf5
    options : dic of (str: str), optional
        Options to pass to the DataFrameReader. Default is no options.

    Returns
    ----------
    sphererdd : SphereRDD instance
        Instance of the Scala SphereRDD class.

    Examples
    ----------
    Load the data
    >>> fn = os.path.join(path_to_conf,
    ...     "../src/test/resources/cartesian_spheres.fits")
    >>> rdd = SphereRDD(spark, fn, "x,y,z,radius",
    ...     False, "fits", {"hdu": "1"})

    Check we have a SphereRDD
    >>> assert("spark3d.spatial3DRDD.SphereRDD" in rdd.toString())

    Count the number of elements
    >>> print(rdd.rawRDD().count())
    20000

    Repartition the data of the RDD (ONION)
    >>> gridtype = "LINEARONIONGRID"
    >>> rdd_part = rdd.spatialPartitioningPython(gridtype,
    ...     rdd.rawRDD().getNumPartitions())

    Repartition the data of the RDD (OCTREE)
    >>> gridtype = "OCTREE"
    >>> rdd_part = rdd.spatialPartitioningPython(gridtype,
    ...     rdd.rawRDD().getNumPartitions())

    Get a RDD with the coordinates of the Point3D
    centers (e.g. useful for plot)
    >>> rdd_centers = rdd.toCenterCoordinateRDDPython(rdd.rawRDD())
    >>> print(round(list(rdd_centers.first())[0], 2))
    0.77

    To see all the available methods:
    >>> print(sorted(rdd.__dir__())) # doctest: +NORMALIZE_WHITESPACE
    ['$lessinit$greater$default$6', '$lessinit$greater$default$7', 'apply',
    'boundary', 'boundary_$eq',
    'com$astrolabsoftware$spark3d$spatial3DRDD$Shape3DRDD$$mapElements$1',
    'equals', 'getClass', 'getDataEnvelope', 'hashCode', 'isSpherical',
    'notify', 'notifyAll', 'partition', 'rawRDD', 'spatialPartitioning',
    'spatialPartitioning$default$2', 'spatialPartitioningPython',
    'spatialPartitioningPython$default$2', 'toCenterCoordinateRDD',
    'toCenterCoordinateRDDPython', 'toString', 'wait']

    """
    scalapath = "com.astrolabsoftware.spark3d.spatial3DRDD.SphereRDD"
    scalaclass = load_from_jvm(scalapath)

    sphererdd = scalaclass(
        spark._jwrapped.sparkSession(), filename, colnames,
        isSpherical, format, options)

    return sphererdd


if __name__ == "__main__":
    """
    Run the doctest using

    python spatial3DRDD.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    from pyspark3d import get_spark_session
    from pyspark3d import load_user_conf

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    dic = load_user_conf()
    spark = get_spark_session(dicconf=dic)

    # Run the test suite
    failure_count, test_count = doctest.testmod(extraglobs={"spark": spark})
    sys.exit(failure_count)



# Copyright 2018 Julien Peloton
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from py4j.java_gateway import JavaObject

from pyspark3d import load_from_jvm
from pyspark3d.converters import scala2python

def windowQuery(rdd: JavaObject, envelope: JavaObject) -> JavaObject:
    """
    Binding around windowQuery (RangeQuery.scala). For full description,
    see `$spark3d/src/main/scala/com/spark3d/spatialOperator/RangeQuery.scala`

    Perform window query, that is match between RDD elements and
    a user-defined window (point, shell, box).

    Parameters
    ----------
    rdd : JavaObject (RDD[A<:Shape3D])
        RDD[A] coming from the Java side. `A` can be Point3D, ShellEnvelope,
        or BoxEnvelope.
    envelope : JavaObject (B<:Shape3D)
        Geometry object coming from the Java side. Can be Point3D,
        ShellEnvelope, or BoxEnvelope.

    Returns
    ----------
    match : JavaObject (RDD[A<:Shape3D])
        RDD[A] containing elements of `rdd` intersecting (within)
        the `envelope`.

    Examples
    ----------
    >>> from pyspark3d.geometryObjects import ShellEnvelope
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import SphereRDD

    Load the data
    >>> fn = os.path.join(path_to_conf,
    ...     "../src/test/resources/cartesian_spheres.fits")
    >>> rdd = SphereRDD(spark, fn, "x,y,z,radius",
    ...     False, "fits", {"hdu": "1"})

    Load the envelope (Sphere at the center, and radius 0.5)
    >>> sh = ShellEnvelope(0.0, 0.0, 0.0, False, 0.0, 0.5)

    Perform the query
    >>> matchRDD = windowQuery(rdd.rawRDD(), sh)
    >>> print("{}/{} objects found in the envelope".format(
    ...     len(matchRDD.collect()), rdd.rawRDD().count()))
    1435/20000 objects found in the envelope
    """
    spark3droot = "com.astrolabsoftware.spark3d."
    scalapath = spark3droot + "spatialOperator.RangeQuery"
    scalaclass = load_from_jvm(scalapath)

    classpath = spark3droot + "python.PythonClassTag.classTagFromObject"
    classtag = load_from_jvm(classpath)

    first_el = rdd.first()
    matchRDD = scalaclass.windowQuery(
        rdd,
        envelope,
        classtag(first_el),
        classtag(envelope))

    return matchRDD

def KNN(
        rdd: JavaObject, queryObject: JavaObject,
        k: int, unique: bool = False) -> list:
    """
    Binding around KNN (SpatialQuery.scala). For full description, see
    `$spark3d/src/main/scala/com/spark3d/spatialOperator/SpatialQuery.scala`

    Finds the K nearest neighbours of the query object within `rdd`.
    The naive implementation here searches through all the the objects in the
    RDD to get the KNN. The nearness of the objects here is decided on the
    basis of the distance between their centers.
    Note that `queryObject` and elements of `rdd` must have the same type
    (either both Point3D, or both ShellEnvelope, or both BoxEnvelope).

    Parameters
    ----------
    rdd : JavaObject (RDD[A<:Shape3D])
        RDD[A] coming from the Java side. `A` can be Point3D, ShellEnvelope,
        or BoxEnvelope.
    envelope : JavaObject (A<:Shape3D)
        Object to which the KNN are to be found. Can be Point3D,
        ShellEnvelope, or BoxEnvelope, but should have the same type as the
        elements of `rdd`.
    k : int
        Number of nearest neighbors requested
    unique : bool, optional
        If True, returns only distinct objects. Default is False.

    Returns
    ----------
    match : List[JavaObject] (List[A<:Shape3D])
        List of the KNN (of type A<:Shape3D).

    Examples
    ----------
    >>> from pyspark3d.geometryObjects import Point3D
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import Point3DRDD

    Load the data (spherical coordinates)
    >>> fn = os.path.join(path_to_conf, "../src/test/resources/astro_obs.fits")
    >>> rdd = Point3DRDD(spark, fn, "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    Define your query point
    >>> pt = Point3D(0.0, 0.0, 0.0, True)

    Perform the query: look for the 5 closest neighbours from `pt`.
    >>> match = KNN(rdd.rawRDD(), pt, 5, True)

    `match` is a list of Point3D. Take the coordinates and convert them
    into cartesian coordinates:
    >>> mod = "com.astrolabsoftware.spark3d.utils.Utils.sphericalToCartesian"
    >>> converter = load_from_jvm(mod)
    >>> match_coord = [converter(m).getCoordinatePython() for m in match]

    Print the distance to the query point (here 0, 0, 0)
    >>> def normit(l: list) -> float: return np.sqrt(sum([e**2 for e in l]))
    >>> print([int(normit(l)*1e6) for l in match_coord])
    [72, 73, 150, 166, 206]
    """
    spark3droot = "com.astrolabsoftware.spark3d."
    scalapath = spark3droot + "spatialOperator.SpatialQuery"
    scalaclass = load_from_jvm(scalapath)

    classpath = spark3droot + "python.PythonClassTag.classTagFromObject"
    classtag = load_from_jvm(classpath)

    match = scalaclass.KNN(rdd, queryObject, k, unique, classtag(queryObject))

    return scala2python(match)

def KNNEfficient(rdd: JavaObject, queryObject: JavaObject, k: int) -> list:
    """
    Binding around KNNEfficient (SpatialQuery.scala). For full description, see
    `$spark3d/src/main/scala/com/spark3d/spatialOperator/SpatialQuery.scala`

    More efficient implementation of the KNN query above.
    First we seek the partitions in which the query object belongs and we
    will look for the knn only in those partitions. After this if the limit k
    is not satisfied, we keep looking similarly in the neighbors of the
    containing partitions.

    Note 1: elements of `rdd` and `queryObject` can have different types
        among Shape3D (Point3D or ShellEnvelope or BoxEnvelope)

    Note 2: KNNEfficient only works on repartitioned RDD (python version).
        See example below.

    Parameters
    ----------
    rdd : JavaObject (RDD[A<:Shape3D])
        Repartitioned RDD[A] coming from the Java side.
        `A` can be Point3D, ShellEnvelope, or BoxEnvelope.
    envelope : JavaObject (B<:Shape3D)
        Object to which the KNN are to be found. `B` can be Point3D,
        ShellEnvelope, or BoxEnvelope, not necessarily the same type as the
        elements of `rdd`.
    k : int
        Number of nearest neighbors requested

    Returns
    ----------
    match : List[JavaObject] (List[A<:Shape3D])
        List of the KNN (of type A<:Shape3D).

    Examples
    ----------
    >>> from pyspark3d.geometryObjects import Point3D
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import Point3DRDD

    Load the raw data (spherical coordinates)
    >>> fn = os.path.join(path_to_conf, "../src/test/resources/astro_obs.fits")
    >>> rdd = Point3DRDD(spark, fn, "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    Repartition the RDD
    >>> rdd_part = rdd.spatialPartitioningPython("LINEARONIONGRID", 5)

    Note that using rdd.spatialPartitioning("LINEARONIONGRID", 5, <classTag>)
    would work as well.

    Define your query point
    >>> pt = Point3D(0.0, 0.0, 0.0, True)

    Perform the query: look for the 5 closest neighbours from `pt`.
    >>> match = KNNEfficient(rdd_part, pt, 5)

    `match` is a list of Point3D. Take the coordinates and convert them
    into cartesian coordinates:
    >>> mod = "com.astrolabsoftware.spark3d.utils.Utils.sphericalToCartesian"
    >>> converter = load_from_jvm(mod)
    >>> match_coord = [converter(m).getCoordinatePython() for m in match]

    Print the distance to the query point (here 0, 0, 0)
    >>> def normit(l: list) -> float: return np.sqrt(sum([e**2 for e in l]))
    >>> print([int(normit(l)*1e6) for l in match_coord])
    [72, 73, 150, 166, 206]
    """
    spark3droot = "com.astrolabsoftware.spark3d."
    scalapath = spark3droot + "spatialOperator.SpatialQuery"
    scalaclass = load_from_jvm(scalapath)

    classpath = spark3droot + "python.PythonClassTag.classTagFromObject"
    classtag = load_from_jvm(classpath)

    el = rdd.first()
    match = scalaclass.KNNEfficient(
        rdd, queryObject, k,
        classtag(el), classtag(queryObject))

    return scala2python(match)

def CrossMatchCenter(
        rddA: JavaObject, rddB: JavaObject,
        epsilon: float, returnType: str) -> JavaObject:
    """
    Binding around CrossMatchCenter (CenterCrossMatch.scala).
    For full description, see
    `$spark3d/src/main/scala/com/spark3d/spatialOperator/CenterCrossMatch.scala`

    Cross match 2 RDD based on the object centers.
    *The two input RDD must have been partitioned the same way!*

    You have to choice to return:
      (1) Elements of (A, B) matching (returnType="AB")
      (2) Elements of A matching B (returnType="A")
      (3) Elements of B matching A (returnType="B")

    Which one you should choose? That depends on what you need:
    (1) gives you all elements but is slow.
    (2) & (3) give you all elements only in one side but is faster.

    Parameters
    ----------
    rddA : JavaObject (RDD[A<:Shape3D])
        RDD whose elements are Shape3D or any extension (Point3D, ...)
    rddB : JavaObject (RDD[B<:Shape3D])
        RDD whose elements are Shape3D or any extension (Point3D, ...)
    epsilon : float
        Tolerance for the distance between 2 centers. Should have the same
        units as the center coordinates.
    returnType : str
        Kind of crossmatch to perform:
            - Elements of (A, B) matching (returnType="AB")
            - Elements of A matching B (returnType="A")
            - Elements of B matching A (returnType="B")

    Returns
    ----------
    rdd : JavaObject
        RDD whose elements depends on `returnType`:
            - Elements of (A, B) matching (returnType="AB")
            - Elements of A matching B (returnType="A")
            - Elements of B matching A (returnType="B")

    Examples
    ----------
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import Point3DRDD

    Load the raw data (spherical coordinates)
    >>> fn = os.path.join(path_to_conf,
    ...     "../src/test/resources/astro_obs_{}_light.fits")
    >>> rddA = Point3DRDD(spark, fn.format("A"), "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})
    >>> rddB = Point3DRDD(spark, fn.format("B"), "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    Apply the same partitioning to both RDD
    >>> rddA_part = rddA.spatialPartitioningPython("LINEARONIONGRID", 100)
    >>> rddB_part = rddB.spatialPartitioningPython(
    ...     rddA_part.partitioner().get())

    Cross match A and B centers and return elements of A matching with B
    >>> matchRDDB = CrossMatchCenter(rddA_part, rddB_part, 0.004, "A")
    >>> print("{}/{} elements in A match with elements of B!".format(
    ...     matchRDDB.count(), rddB_part.count()))
    19/1000 elements in A match with elements of B!
    """
    spark3droot = "com.astrolabsoftware.spark3d."
    scalapath = spark3droot + "spatialOperator.CenterCrossMatch"
    scalaclass = load_from_jvm(scalapath)

    classpath = spark3droot + "python.PythonClassTag.classTagFromObject"
    classtag = load_from_jvm(classpath)

    elA = rddA.first()
    elB = rddB.first()
    matchRDD = scalaclass.CrossMatchCenter(
        rddA,
        rddB,
        epsilon,
        returnType,
        classtag(elA),
        classtag(elB))

    return matchRDD

def CrossMatchHealpixIndex(
        rddA: JavaObject, rddB: JavaObject,
        nside: int, returnType: str) -> JavaObject:
    """
    Binding around CrossMatchHealpixIndex (PixelCrossMatch.scala).
    For full description, see
    `$spark3d/src/main/scala/com/spark3d/spatialOperator/PixelCrossMatch.scala`

    Cross match 2 RDD based on the healpix index of geometry center.
    The cross-match is done partition-by-partition, which means the two
    RDD must have been partitioned by the same partitioner.

    You have to choice to return:
      (1) Elements of (A, B) matching (returnType="AB")
      (2) Elements of A matching B (returnType="A")
      (3) Elements of B matching A (returnType="B")
      (4) Healpix pixel indices matching (returnType="healpix")

    Which one you should choose? That depends on what you need:
    (1) gives you all elements but is slow.
    (2) & (3) give you all elements only in one side but is faster.
    (4) gives you only healpix center but is even faster.

    Parameters
    ----------
    rddA : JavaObject (RDD[A<:Shape3D])
        RDD whose elements are Shape3D or any extension (Point3D, ...)
    rddB : JavaObject (RDD[B<:Shape3D])
        RDD whose elements are Shape3D or any extension (Point3D, ...)
    nside : int
        Resolution of the underlying healpix map used to convert angle
        coordinates to healpix index.
    returnType : str
        Kind of crossmatch to perform:
            - Elements of (A, B) matching (returnType="AB")
            - Elements of A matching B (returnType="A")
            - Elements of B matching A (returnType="B")

    Returns
    ----------
    rdd : JavaObject
        RDD whose elements depends on `returnType`:
            - Elements of (A, B) matching (returnType="AB")
            - Elements of A matching B (returnType="A")
            - Elements of B matching A (returnType="B")
            - Healpix pixel indices matching (returnType="healpix")

    Examples
    ----------
    >>> from pyspark3d_conf import path_to_conf
    >>> from pyspark3d.spatial3DRDD import Point3DRDD

    Load the raw data (spherical coordinates)
    >>> fn = os.path.join(path_to_conf,
    ...     "../src/test/resources/astro_obs_{}_light.fits")
    >>> rddA = Point3DRDD(spark, fn.format("A"), "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})
    >>> rddB = Point3DRDD(spark, fn.format("B"), "Z_COSMO,RA,DEC",
    ...     True, "fits", {"hdu": "1"})

    Apply the same partitioning to both RDD
    >>> rddA_part = rddA.spatialPartitioningPython("LINEARONIONGRID", 100)
    >>> rddB_part = rddB.spatialPartitioningPython(
    ...     rddA_part.partitioner().get())

    Cross match A and B centers and return elements of A matching with B
    >>> matchRDDB = CrossMatchHealpixIndex(rddA_part, rddB_part, 512, "A")
    >>> print("{}/{} elements in A match with elements of B!".format(
    ...     matchRDDB.count(), rddB_part.count()))
    15/1000 elements in A match with elements of B!
    """
    spark3droot = "com.astrolabsoftware.spark3d."
    scalapath = spark3droot + "spatialOperator.PixelCrossMatch"
    scalaclass = load_from_jvm(scalapath)

    classpath = spark3droot + "python.PythonClassTag.classTagFromObject"
    classtag = load_from_jvm(classpath)

    elA = rddA.first()
    elB = rddB.first()
    matchRDD = scalaclass.CrossMatchHealpixIndex(
        rddA,
        rddB,
        nside,
        returnType,
        classtag(elA),
        classtag(elB))

    return matchRDD


if __name__ == "__main__":
    """
    Run the doctest using

    python spatialOperator.py

    If the tests are OK, the script should exit gracefuly, otherwise the
    failure(s) will be printed out.
    """
    import sys
    import doctest
    import numpy as np

    # Numpy introduced non-backward compatible change from v1.14.
    if np.__version__ >= "1.14.0":
        np.set_printoptions(legacy="1.13")

    from pyspark3d import get_spark_session
    from pyspark3d import load_user_conf

    # Load the spark3D JAR+deps, and initialise the spark session.
    # In a pyspark shell, you do not need this.
    dic = load_user_conf()
    spark = get_spark_session(dicconf=dic)

    # Run the test suite
    failure_count, test_count = doctest.testmod(extraglobs={"spark": spark})
    sys.exit(failure_count)
